'''
The functions in this module support the ability to calculate stats for
censored data. These functions require (i) a CensoredData input object that
contains column names to be used for the analysis, (ii) a DataFrame that
has been validated and preprocessed, (iii) the specific column grouping to
use to combine data, and (iv) additional stat specific inputs which currently
only includes percentile and method for the percentile function.

These functions all combine intervals within a group to a single interval
that represents the potential range of the desired stat based on all possible
values of the original intervals.

Functions include:
    -maximum
    -minimum
    -mean/average
    -addition
    -percentile/median

'''

import numpy as np


#%% Extrema stats (Maximum/Minimum)

def _extrema_interval(cdf, data, groupby_cols, stat_name, stat='max'):
    
    # Create a copy of the data
    df = data.copy()
    
    # Create stat column
    df[cdf.stat_col] = stat_name
    
    # Create empty list if no groups
    if groupby_cols == None:
        groupby_cols = []
    
    # Determine the bounds and boundaries for the maximum (or minimum)
    
    # Determine whether the left bound is open or closed by comparing the
    # largest (or least) open bound with the largest (or least) closed bound
    # Repeat for the right bound
    left = df.groupby(groupby_cols + [cdf.stat_col, cdf.left_boundary_col])
    right = df.groupby(groupby_cols + [cdf.stat_col, cdf.right_boundary_col])
    
    # Only keep the maximum (or minimum) for each group
    if stat == 'max':
        left = left[cdf.left_bound_col].max()
        right = right[cdf.right_bound_col].max()
    elif stat == 'min':
        left = left[cdf.left_bound_col].min()
        right = right[cdf.right_bound_col].min()
    else:
        raise ValueError('Wrong input for "stat" in _extrema_interval()')
    
    # Unstack the boundaries into their own columns
    left = left.unstack(cdf.left_boundary_col)
    right = right.unstack(cdf.right_boundary_col)
    
    # Ensure both boundary column exists
    for frame in [left, right]:
        for boundary in ['Open', 'Closed']:
            if boundary not in frame.columns:
                frame[boundary] = np.nan
    
    # Determine conditions where the boundary is Closed
    # For maximum stat,
    if stat == 'max':
        # For left bound, use maximum closed bound if it is larger than
        # the largest Open boundary.
        left_condition = (left['Closed'] > left['Open'])
        # For right bound, use maximum closed bound if it is larger than
        # or equal to the largest Open boundary.
        right_condition = (right['Closed'] >= right['Open'])
    # For minimum stat
    else:
        # For left bound, use minimum closed bound if it is less than
        # or equal to least Open boundary.
        left_condition = (left['Closed'] <= left['Open'])
        # For right bound, use minimum closed bound if it is less than
        # the least Open boundary.
        right_condition = (right['Closed'] < right['Open'])
    
    # Always Closed boundary if there is no Open boundary
    left_condition |= left['Open'].isna()
    right_condition |= right['Open'].isna()
    
    # Set left bounds and boundaries based on the conditions
    left[cdf.left_boundary_col] = (
        np.where(left_condition,'Closed','Open'))
    
    left[cdf.left_bound_col] = (
        np.where(left_condition,left['Closed'],left['Open']))
    
    left = left[[cdf.left_boundary_col, cdf.left_bound_col]]
    
    # Set right bounds and boundaries based on the conditions
    right[cdf.right_bound_col] = (
        np.where(right_condition,right['Closed'],right['Open']))
    
    right[cdf.right_boundary_col] = (
        np.where(right_condition,'Closed','Open'))
    
    right = right[[cdf.right_bound_col, cdf.right_boundary_col]]
    
    # Merge the two dataframes to create the full interval
    # Check the merge is 1-to-1
    df = left.merge(right,
                    how = 'outer',
                    on = groupby_cols + [cdf.stat_col],
                    validate = '1:1')
    
    # Reset index
    df = df.reset_index()
    
    return df
    
def _maximum_interval(cdf, data, groupby_cols, stat_name='Maximum'):
    
    return _extrema_interval(cdf, data, groupby_cols, stat_name, stat='max')

def _minimum_interval(cdf, data, groupby_cols, stat_name='Minimum'):
    
    return _extrema_interval(cdf, data, groupby_cols, stat_name, stat='min')

#%% Addition based stats (Mean/Average/Sum)

def _mean_or_sum_interval(cdf, data, groupby_cols, stat_name, stat='mean'):
    
    # Create a copy of the data
    df = data.copy()
    
    # Create stat column
    df[cdf.stat_col] = stat_name
    
    # Check stat name
    if stat not in ['mean', 'sum']:
        raise ValueError('Wrong input for "stat" in _mean_or_sum_interval()')
    
    # Create empty list if no groups
    if groupby_cols == None:
        groupby_cols = []
    
    # Change notation of 'Closed' and 'Open' boundaries to 0 and 1, 
    # respectively. The presence of any open boundaries on one side ensures
    # that the interval for the result is also open on that side
    df[[cdf.left_boundary_col, cdf.right_boundary_col]] = (
        df[[cdf.left_boundary_col, cdf.right_boundary_col]]
            .replace(['Closed','Open'], [0,1]))
    
    # Get the left/right bounds of the result by averaging or adding all 
    # left/right bounds within the group.
    # Determine whether any Open (now value of 1) boundaries exist by
    # taking max of boundary values. If there are any open boundaries used in 
    # the bound mean/sum, then the resulting mean/sum should be open
    # Determine the min and max of the bounds to correct situations where
    # the bound is infinite
    df = (
        df.groupby(groupby_cols + [cdf.stat_col])
            .agg(**{
                cdf.left_boundary_col: (cdf.left_boundary_col, 'max'),
                '__Minimum__': (cdf.left_bound_col, 'min'),
                cdf.left_bound_col: (cdf.left_bound_col, stat),
                cdf.right_bound_col: (cdf.right_bound_col, stat),
                '__Maximum__': (cdf.right_bound_col, 'max'),
                cdf.right_boundary_col: (cdf.right_boundary_col, 'max')
                })
        )
    
    # Replace integers with text for boundaries
    df[[cdf.left_boundary_col, cdf.right_boundary_col]] = (
        df[[cdf.left_boundary_col, cdf.right_boundary_col]]
            .replace([0,1], ['Closed','Open']))
    
    # Means/sums with infinite values produce nan values rather than 
    # np.inf values. Convert nan to inf only if infinite values are
    # included in the mean/sum
    df[cdf.left_bound_col] = (
        np.where(df['__Minimum__'] == -np.inf,
                 -np.inf,
                 df[cdf.left_bound_col]
                 )
        )
    df[cdf.right_bound_col] = (
        np.where(df['__Maximum__'] == np.inf,
                 np.inf,
                 df[cdf.right_bound_col]
                 )
        )
    
    # Drop working columns
    df = df.drop(df.filter(regex='__').columns, axis=1)
    
    # Reset index
    df = df.reset_index()
    
    return df

def _mean_interval(cdf, data, groupby_cols, stat_name='Mean'):
    
    return _mean_or_sum_interval(cdf, data, groupby_cols,
                                 stat_name, stat='mean')

def _sum_interval(cdf, data, groupby_cols, stat_name='Sum'):
    
    return _mean_or_sum_interval(cdf, data, groupby_cols,
                                 stat_name, stat='sum')

#%% Percentile stats (Percentile/Median)

def _percentile_interval(cdf,
                         data,
                         percentile,
                         groupby_cols,
                         stat_name = 'Percentile',
                         method = 'hazen'):
    
    # Create a copy of the data
    df = data.copy()
    
    # Create stat column
    df[cdf.stat_col] = stat_name
    
    # Create empty list if no groups
    if groupby_cols == None:
        groupby_cols = []
    
    # Set values for percentile methods
    method_dict = {'weiball': 0.0,
                   'tukey': 1/3,
                   'blom': 3/8,
                   'hazen': 1/2,
                   'excel':1.0}
    # https://en.wikipedia.org/wiki/Percentile
    C = method_dict[method]
    
    # Convert percentile to be between 0 and 1
    percentile = percentile/100
    
    # Create column for the size of each group
    df['__Size__'] = (df.groupby(groupby_cols + [cdf.stat_col])
                        .transform('size'))
    
    # Determine the rank in each group for the percentile
    # Use rounding at 8 decimals to prevent artificial decimals
    df['__Rank__'] = round(C + percentile*(df['__Size__'] + 1 - 2*C), 8)
    
    # Create warning column to flag if the percentile rank is outside
    # the range of possible values. If rank is less than 1, the minimum will
    # be returned for the percentile; if the rank is greater than the size
    # of the group, the maximum will be returned.
    df[cdf.warning_col] = ''
    max_condition = (df['__Rank__'] > df['__Size__'])
    min_condition = (df['__Rank__'] < 1)
    df.loc[max_condition, cdf.warning_col] = '(low count, used max)'
    df.loc[min_condition, cdf.warning_col] = '(low count, used min)'
    df.loc[max_condition, '__Rank__'] = df['__Size__']
    df.loc[min_condition, '__Rank__'] = 1
    
    # For the left bound, change notation of Closed and Open boundaries to 0
    # and 1, respectively. Use 0 for Closed to ensure that Closed boundaries
    # are sorted smaller than Open boundaries when the left bound is tied
    df[cdf.left_boundary_col] = (
        df[cdf.left_boundary_col].replace(['Closed','Open'], [0,1]))
    
    # For the right bound, change notation of Closed and Open boundaries to 1
    # and 0, respectively. Use 1 for Closed to ensure that Closed boundaries
    # are sorted larger than Open boundaries when the right bound is tied
    df[cdf.right_boundary_col] = (
        df[cdf.right_boundary_col].replace(['Closed','Open'], [1,0]))
    
    # Sort the left and right bounds
    # Reduce working dataframes to relevant columns
    left = (
        df.copy()[groupby_cols + [cdf.stat_col] +
                  ['__Size__', '__Rank__', cdf.warning_col] +
                  [cdf.left_boundary_col, cdf.left_bound_col]]
            .sort_values(
                by = [cdf.left_bound_col, cdf.left_boundary_col]
                )
        )
    right = (
        df.copy()[groupby_cols + [cdf.stat_col] +
                  ['__Size__', '__Rank__', cdf.warning_col] +
                  [cdf.right_boundary_col, cdf.right_bound_col]]
            .sort_values(
                by = [cdf.right_bound_col, cdf.right_boundary_col]
                )
        )
    
    # Add index for each group
    left['__Index__'] = (left.groupby(groupby_cols + [cdf.stat_col])
                             .cumcount() + 1)
    right['__Index__'] = (right.groupby(groupby_cols + [cdf.stat_col])
                               .cumcount() + 1)
    
    def find_values_within_one_of_rank(df):
        
        # Determine proximity of each result to percentile rank using the index
        
        # Set default as 0
        df['__Proximity__'] = 0
        
        condition_set = [
            # If the percentile rank is a whole number,
            # then use that index result
            (df['__Rank__'] == df['__Index__']),
            # If the percentile rank is less than 1 above the index value,
            # then assign the appropriate contribution to that index value
            ((df['__Rank__'] - df['__Index__'])
                                         .between(0,1,inclusive='neither')),
            # If the percentile rank is less than 1 below the index value,
            # then assign the appropriate contribution to that index value
            ((df['__Index__'] - df['__Rank__'])
                                         .between(0,1,inclusive='neither'))
            ]
        
        proximities = [
            1,
            1 - (df['__Rank__'] - df['__Index__']),
            1 - (df['__Index__'] - df['__Rank__'])
            ]
        
        # Set proximity of each result to rank
        df['__Proximity__'] = np.select(condition_set, proximities, np.nan)
        
        # Drop non-contributing rows
        df = df[df['__Proximity__'] > 0]
        
        return df
    
    left = find_values_within_one_of_rank(left)
    right = find_values_within_one_of_rank(right)
    
    # Calculate contribution for using proximity to rank
    left['__Contribution__'] = (
        left['__Proximity__'] * left[cdf.left_bound_col])
    right['__Contribution__'] = (
        right['__Proximity__'] * right[cdf.right_bound_col])
    
    # Determine bound and boundary using the sum of the contributions
    # and an open boundary if any of the contributing values is open
    # Replace nan sums with infinite bound
    left = (
        left.groupby(groupby_cols + [cdf.stat_col, cdf.warning_col])
            .agg(**{
                cdf.left_boundary_col: (cdf.left_boundary_col, 'max'),
                cdf.left_bound_col: ('__Contribution__', 'sum'),
                '__Minimum__': ('__Contribution__', 'min')
                }))
    right = (
        right.groupby(groupby_cols + [cdf.stat_col, cdf.warning_col])
            .agg(**{
                cdf.right_bound_col: ('__Contribution__', 'sum'),
                cdf.right_boundary_col: (cdf.right_boundary_col, 'min'),
                '__Maximum__': ('__Contribution__', 'max')
                }))
    
    # Replace the numeric value for the boundary
    left[cdf.left_boundary_col] = (
        left[cdf.left_boundary_col].replace([0,1], ['Closed','Open']))
    right[cdf.right_boundary_col] = (
        right[cdf.right_boundary_col].replace([1,0], ['Closed','Open']))
    
    # Replace nan bounds with infinite bound
    left[cdf.left_bound_col] = (
        np.where(left['__Minimum__'] == -np.inf, -np.inf,
                 left[cdf.left_bound_col]))
    right[cdf.right_bound_col] = (
        np.where(right['__Maximum__'] == np.inf, np.inf,
                 right[cdf.right_bound_col]))
    
    
    # Merge the two boundaries to create the interval for the percentile
    # Check that the merge is 1-to-1
    df = left.merge(right,
                    how = 'outer',
                    on = groupby_cols + [cdf.stat_col, cdf.warning_col],
                    validate = '1:1'
                    )
    
    # Drop working columns
    df = df.drop(df.filter(regex='__').columns, axis=1)
    
    # Reset index
    df = df.reset_index()
    
    return df

def _median_interval(cdf, data, groupby_cols, stat_name='Median'):
    
    return _percentile_interval(cdf, data, 50, groupby_cols, stat_name)


'''
The function in this module support the ability to merge count information
to stat tables. Required information includes (i) a CensoredData input object
that contains the original data table, (ii) a DataFrame that contains the
final stat outputs, (iii) the specific column groupings used to combine data,
and (iv) the names to give the count for each grouping.

'''

def _merge_count_info(cdf, stat_data, groupby_cols, count_cols, filters):
    
    # Create a copy of the data
    df = cdf.data.copy()
    
    # Apply filters
    if filters:
        for col_name in filters:
            df = df[df[col_name].isin(filters[col_name])]
    
    # Validate count_cols
    _validate_count_cols(cdf, groupby_cols, count_cols)
    
    # If groupby_cols is None
    if groupby_cols == None:
        stat_data[count_cols[0]] = len(df)
    
    else:
        # Loop through list of groupings and sum counts
        for i in range(len(groupby_cols)):
            df[count_cols[i]] = 1
            df = (df.groupby(groupby_cols[i])[count_cols[:i+1]]
                      .sum()
                      .reset_index())
        # Merge count info to stat data using last grouping
        stat_data = stat_data.merge(df,
                                    how='outer',
                                    on=groupby_cols[-1],
                                    validate = '1:1')
    
    return stat_data


def _validate_count_cols(cdf, groupby_cols, count_cols):
    
    # Check that count_cols is a list
    if not isinstance(count_cols, list):
        raise ValueError('count_cols must be a list. '
            f'{count_cols=} was passed.')
    
    # Check that items are all strings
    if not (all(isinstance(name, str) for name in count_cols)):
        raise ValueError('count_cols must be a list of text strings. '
            f'Non-text names were detected in {count_cols=}')
    
    # Check that names are unique
    if (len(count_cols) != len(set(count_cols))):
        raise ValueError('count_cols must contain unique names. '
            f'Non-unique names were detected in {count_cols=}')
    
    # Check that single name provided if groupby_cols is None or []
    if groupby_cols == None:
        if len(count_cols) != 1:
            raise ValueError('A single column name should be provided as '
                'a list when groupby_cols is not used. '
                f'{count_cols=} was passed.')
    else:
        # Check that lists are same length if groupby_cols is used
        if len(count_cols) != len(groupby_cols):
            raise ValueError('count_cols must have the same number of '
                'column names as there are grouping lists in groupby_cols. '
                f'{len(count_cols)} names were provided in count_cols and '
                f'{len(groupby_cols)} names were provided in groupby_cols.')
        
        # Check names are not used in the groupby_cols
        groupby_names = sum(groupby_cols, [])
        common_names = [name for name in count_cols if name in groupby_names]
        if len(common_names) != 0:
            raise ValueError('count_cols cannot include names that already '
                'exist in the data which are also used for grouping. '
                f'Repeated names include: {common_names}')


'''
These functions apply a specific rounding algorithm based on value. Some values
are rounded based on specified decimal places while others are based on
significant digits.

'''

import numpy as np

def _string_precision(value,
                      thousands_comma=False):
    '''
    A function that applies a specified rounding method that is value
    dependent. This method is specifically designed to reflect the
    typical measurement precision for water quality results. Depending on the
    value, the rounding is either to a particular number of decimal places or
    to a particular number of significant digits.

    Parameters
    ----------
    value : float
        Numeric value that will be rounded.

    Returns
    -------
    string : string
        The rounded value expressed as a string to the appropriate precision.

    '''
    
    # Check for infinite values
    if abs(value) == np.inf:
        string = str(value)
    # Values above 100 or are rounded to 100 should be rounded to 3 significant digits
    elif round(abs(value),1) >= 100:
        string = f'{value:.3g}'
        # Include thousands separator, depending on input
        if thousands_comma:
            string = f'{int(float(string)):,}'
        else:
            string = str(int(float(string)))
    # Values above 10 or are rounded to 10 should be rounded to 1 decimal place
    elif round(abs(value),2) >= 10:
        string = f'{value:.1f}'
    # Values above 0.2 or are rounded to 0.2 should be rounded to 2 decimal places
    elif round(abs(value),3) >= 0.2:
        string = f'{value:.2f}'
    # Values above 0.1 or are rounded to 0.1 should be rounded to 3 decimal places
    elif round(abs(value),3) >= 0.1:
        string = f'{value:.3f}'
    # Values below 0.1 should be rounded to 2 significant digits
    else:
        string = f'{value:.2g}'

    return string


def _numeric_precision(value):
    '''
    A function that returns a float data type from a rounding function instead
    of a string data type.

    Parameters
    ----------
    value : float
        Float value that may have more decimal places or significant digits
        than is appropriate

    Returns
    -------
    float
        Float value that is rounded appropriately

    '''
    
    # Return the same rounded result as string_precision but as a float
    return float(_string_precision(value))




'''
These functions validate user input values.

'''

import pandas as pd

#%% Validate CensoredData

def _validate_cdf(cdf):
    
    #%% Check data input
    
    # Check that supplied data is a DataFrame
    if not isinstance(cdf.data, pd.core.frame.DataFrame):
        raise ValueError('The data supplied to CensoredData() must be a '
            'pandas DataFrame. Instead, an object was passed with type: '
            f'{type(cdf.data).__name__}')
    
    # Ensure that provided data is a copy and not indexed
    cdf.data = cdf.data.copy().reset_index()
    
    #%% Check value_col exists as column in data
    
    if cdf.value_col not in cdf.data.columns:
        raise ValueError('The value column supplied to CensoredData() '
            'was not found as a column in the data. Columns include '
            f'{cdf.data.columns.to_list()} which does not include '
            f'{repr(cdf.value_col)}.')
    
    #%% Check the four stat settings
    
    boolean_settings = ['include_negative_interval',
                        'focus_high_potential',
                        'precision_rounding',
                        'thousands_comma',
                        'output_interval']
    
    for setting in boolean_settings:
        setting_mode = getattr(cdf, setting)
        if not isinstance(setting_mode, bool):
            raise ValueError(f'The value supplied to {setting} must be '
                'True or False. Instead, the value supplied was '
                f'{repr(setting_mode)}.')
    
    # True or False are considered integer types so check for boolean too
    if ((isinstance(cdf.precision_tolerance_to_drop_censor, bool)) |
        (not isinstance(cdf.precision_tolerance_to_drop_censor,
                       (int, float)))):
        raise ValueError(f'The value supplied to '
            'precision_tolerance_to_drop_censor must be numeric. Instead, '
            'the value supplied was '
            f'{repr(cdf.precision_tolerance_to_drop_censor)}.')
    
    #%% Check column names for conflicts
    
    # Check that columns created within methods don't conflict with column
    # names in provided data
    # Check that new column names are string values
    
    # Create list of column names that may be created
    created_columns = [attribute for attribute in cdf.__annotations__
                               if attribute.endswith('_col') and
                                   attribute != 'value_col']
    
    # Check the list for duplicates
    if len(created_columns) != len(set(created_columns)):
        raise ValueError('Two default column names have been set to the '
            'same name. Check that assigned values are unique and not '
            'identical to any default values.')
    
    
    for col in created_columns:
        col_name = getattr(cdf, col)
        if (col_name in cdf.data.columns) & (col_name != cdf.value_col):
            raise ValueError('The data contains a column named '
                f'"{col_name}" which may be used in the output. Either '
                'rename this column or provide an alternative value for '
                f'{col} when initialising CensoredDataFrame.')
        
        if not isinstance(col_name, str):
            raise ValueError('Text values are required for column names. '
                f'The column name {col_name} should be changed to a text '
                f'value for {col}.')
    
    # Check that the data has no columns starting with double underscore
    dunder_cols = [col for col in cdf.data if col.startswith('__')]
    if len(dunder_cols) != 0:
        raise ValueError('Columns starting with "__" are used within '
            'the stat methods. Rename the following columns: '
            f'{dunder_cols}')
    
    #%% Check that there are no nan values or empty strings
    if ((cdf.data[cdf.value_col].isnull().any()) |
        (cdf.data[cdf.value_col].astype(str).str.len().min() == 0)):
        raise ValueError('Missing values need to be removed from the data '
            'before it can be analysed.')

#%% Groupby validation

def _validate_groupby_cols(cdf, groupby_cols):
    
    # Skip None case
    if groupby_cols == None:
        pass
    else:
        # Check that groupby_cols is a nested list
        if ((not isinstance(groupby_cols, list)) | 
            (not (all(isinstance(group, list) for group in groupby_cols)))):
            raise ValueError('groupby_cols needs to be a nested list. '
                f'The provided input was {groupby_cols}')
        
        # Check that column names are included in data
        if not (all(name in cdf.data.columns for name in groupby_cols[0])):
            raise ValueError('Columns used in groupby_cols were not found in '
                'the DataFrame. Names not found include: '
                f'{[x for x in groupby_cols[0] if x not in cdf.data.columns]}')
        
        # Check that subsequent lists are subsets
        for i in range(len(groupby_cols)-1):
            # Create list of columns not in prior group
            errors = [x for x in groupby_cols[i+1] if x not in groupby_cols[i]]
            if len(errors) != 0:
                raise ValueError('When using tiered groupings, subsequent '
                    'groupings must be subsets of earlier groups. The '
                    'following column names were not included in all earlier '
                    f'groups: {errors}')
        
        # Check for null values in groupby columns
        if cdf.data[groupby_cols[0]].isnull().values.any():
            raise ValueError('Null values found in one of the columns used '
                'for grouping.')
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            
            



import numpy as np
import pandas as pd
from dataclasses import dataclass

from censoredsummarystats.validation import (
    _validate_cdf,
    _validate_groupby_cols)
from censoredsummarystats.stat_interval_aggregation import (
    _maximum_interval,
    _minimum_interval,
    _mean_interval,
    _sum_interval,
    _percentile_interval,
    _median_interval)
from censoredsummarystats.interval_to_result import _interval_to_result
from censoredsummarystats.percent_exceedance import (
    _determine_exceedances,
    _group_exceedances,
    _percent_exceedances)
from censoredsummarystats.merge_count_info import _merge_count_info

@dataclass
class CensoredData:
    data: pd.core.frame.DataFrame
    value_col: any
    include_negative_interval: bool = False
    focus_high_potential: bool = True
    precision_tolerance_to_drop_censor: float = 0.25
    precision_rounding: bool = True
    thousands_comma: bool = False
    output_interval: bool = True
    stat_col: str = 'Statistic'
    result_col: str = 'Result'
    censor_col: str = 'CensorComponent'
    numeric_col: str = 'NumericComponent'
    left_boundary_col: str = 'LeftBoundary'
    left_bound_col: str = 'LeftBound'
    right_bound_col: str = 'RightBound'
    right_boundary_col: str = 'RightBoundary'
    interval_col: str = 'Interval'
    threshold_col: str = 'Threshold'
    exceedances_col: str = 'Exceedances'
    non_exceedances_col: str = 'NonExceedances'
    ignored_col: str = 'IgnoredValues'
    warning_col: str = 'Warning'
    
    def __post_init__(self):
        
        #%% Check data input
        
        _validate_cdf(self)
        
        #%% Split value into censor/numeric components
        
        # Create temporary column for first character
        first_character = self.data[self.value_col].astype(str).str[0]
        # Determine if the value has a censor component
        is_censored = first_character.isin(['<','≤','≥','>'])
        
        # Create censor column using first characters that are censors
        self.data[self.censor_col] = (
            np.where(
                is_censored,
                first_character,
                ''
                )
            )
        
        # Create numeric column using left remaining part of value
        self.data[self.numeric_col] = (
            np.where(
                is_censored,
                self.data[self.value_col].astype(str).str[1:],
                self.data[self.value_col]
                )
            )
        
        # Convert numeric component to float
        try:
            self.data[self.numeric_col] = (self.data[self.numeric_col]
                                               .astype(float))
        except ValueError:
            raise Exception('At least one value could not be converted to a '
                'numeric data type. Check that each value has no more than a '
                'single censor character that is one of the following: '
                '<, ≤, ≥, >')
            
        
        #%% Create interval for value (left/right bounds)
        
        # Define where the left bound is closed
        self.data[self.left_boundary_col] = (
                np.where(self.data[self.censor_col].isin(['','≥']),
                          'Closed',
                          'Open'
                          )
                )
        
        # Define where the left bound is unlimited
        self.data[self.left_bound_col] = (
                np.where(self.data[self.censor_col].isin(['<','≤']),
                                    -np.inf,
                                    self.data[self.numeric_col]
                                    )
                )
        
        # Define where the right bound is unlimited
        self.data[self.right_bound_col] = (
                np.where(self.data[self.censor_col].isin(['≥','>']),
                                    np.inf,
                                    self.data[self.numeric_col]
                                    )
                )
        
        # Define where the right bound is closed
        self.data[self.right_boundary_col] = (
                np.where(self.data[self.censor_col].isin(['≤','']),
                          'Closed',
                          'Open'
                          )
                )
        
        # If left censored are assumed positive
        if not self.include_negative_interval:
            # Check for any negative values (exclude -inf)
            if (self.data[self.left_bound_col]
                    .between(-np.inf, 0, inclusive='neither')
                    .any()
                    ):
                raise ValueError('Negative values exist in the data. Resolve '
                    'negative values or set include_negative_interval = True')
            # Set any -inf left bounds to 0 with closed boundary
            condition = (self.data[self.left_bound_col] < 0)
            self.data[self.left_boundary_col] = (
                np.where(condition,
                          'Closed',
                          self.data[self.left_boundary_col]
                          )
                )
            self.data[self.left_bound_col] = (
                np.where(condition,
                          0.0,
                          self.data[self.left_bound_col]
                          )
                )
    
    #%% Statistic methods
    
    def _general_stat(self,
                      stat,
                      groupby_cols = None,
                      count_cols = None,
                      stat_name = None,
                      filters = None):
        
        # Validate groupby_cols
        _validate_groupby_cols(self, groupby_cols)
        
        # Create a copy of the CensoredData object
        df = self.data.copy()
        
        # Apply filters
        if filters:
            for col_name in filters:
                df = df[df[col_name].isin(filters[col_name])]
        
        # If no groupby_cols provided, calculate stat for full dataset
        if groupby_cols == None:
            df = stat(self, df, None, stat_name)
        else:
            for grouping in groupby_cols:
                df = stat(self, df, grouping, stat_name)
        
        # Set focus of high or low
        if stat == _maximum_interval:
            focus_high_potential = True
        elif stat == _minimum_interval:
            focus_high_potential = False
        else:
            focus_high_potential = self.focus_high_potential
        
        # Create result from interval information
        df = _interval_to_result(self, df, focus_high_potential)
        
        # Merge count info
        if count_cols:
            df = _merge_count_info(self, df, groupby_cols, count_cols, filters)
        
        return df
    
    def maximum(self,
                groupby_cols = None,
                count_cols = None,
                stat_name = 'Maximum',
                filters = None):
        
        return self._general_stat(_maximum_interval,
                                  groupby_cols,
                                  count_cols,
                                  stat_name,
                                  filters)
    
    def minimum(self,
                groupby_cols = None,
                count_cols = None,
                stat_name = 'Minimum',
                filters = None):
        
        return self._general_stat(_minimum_interval,
                                  groupby_cols,
                                  count_cols,
                                  stat_name,
                                  filters)
    
    def mean(self,
             groupby_cols = None,
             count_cols = None,
             stat_name = 'Mean',
             filters = None):
        
        return self._general_stat(_mean_interval,
                                  groupby_cols,
                                  count_cols,
                                  stat_name,
                                  filters)
    
    def average(self,
                groupby_cols = None,
                count_cols = None,
                stat_name = 'Average',
                filters = None):
        
        return self._general_stat(_mean_interval,
                                  groupby_cols,
                                  count_cols,
                                  stat_name,
                                  filters)
    
    def add(self,
            groupby_cols = None,
            count_cols = None,
            stat_name = 'Sum',
            filters = None):
        
        return self._general_stat(_sum_interval,
                                  groupby_cols,
                                  count_cols,
                                  stat_name,
                                  filters)
    
    def median(self,
               groupby_cols = None,
               count_cols = None,
               stat_name = 'Median',
               filters = None):
        
        return self._general_stat(_median_interval,
                                  groupby_cols,
                                  count_cols,
                                  stat_name,
                                  filters)
        
    def percentile(self,
                   percentile,
                   groupby_cols = None,
                   count_cols = None,
                   stat_name = 'Percentile',
                   method = 'hazen',
                   filters = None):
        
        # Create a copy of the CensoredData object
        df = self.data.copy()
        
        # Apply filters
        if filters:
            for col_name in filters:
                df = df[df[col_name].isin(filters[col_name])]
        
        # Validate percentile
        if (percentile > 100) | (percentile < 0):
            raise ValueError('The percentile must be between 0 and 100.')
        
        # Validate groupby_cols
        _validate_groupby_cols(self, groupby_cols)
        
        # If no groupby_cols provided, calculate percentile for full dataset
        if groupby_cols == None:
            df = _percentile_interval(self, df, percentile,
                                      None, stat_name, method)
        else:
            # Use median for all but the last group
            for grouping in groupby_cols:
                if grouping == groupby_cols[-1]:
                    df = _percentile_interval(self, df, percentile,
                                              grouping, stat_name, method)
                else:
                    df = _median_interval(self, df, grouping, stat_name)
        
        # Create result from interval information
        df = _interval_to_result(self, df)
        
        # Append warning message to result
        df[self.result_col] += ' ' + df[self.warning_col]
        
        # Drop warning column
        df = df.drop(columns=[self.warning_col])
        
        # Merge count info
        if count_cols:
            df = _merge_count_info(self, df, groupby_cols, count_cols, filters)
        
        return df
    
    def percent_exceedance(self,
                           threshold,
                           threshold_is_exceedance,
                           groupby_cols = None,
                           count_cols = None,
                           stat_name = 'Percent Exceedance',
                           round_to = 2,
                           filters = None):
        
        # Create a copy of the CensoredData object
        df = self.data.copy()
        
        # Apply filters
        if filters:
            for col_name in filters:
                df = df[df[col_name].isin(filters[col_name])]
        
        # Validate groupby_cols
        _validate_groupby_cols(self, groupby_cols)
        
        # Validate round_to input
        if (isinstance(round_to, bool) |
                (not isinstance(round_to, int))):
            raise ValueError(f'The value supplied to round_to must be an '
                'integer. Instead, the value supplied was '
                f'{repr(round_to)}.')
        
        # Assess the data for exceedances
        df = _determine_exceedances(self, df, threshold, threshold_is_exceedance)
        
        # Create stat column
        df[self.stat_col] = stat_name
        
        # Create empty list if no groups
        if groupby_cols == None:
            groupby_cols = []
        else:
            if len(groupby_cols) > 2:
                raise ValueError('The percent exceedance statistic can only '
                    'receive two lists for groupby_cols yet '
                    f'{len(groupby_cols)} were passed.')
        
        if len(groupby_cols) == 2:
            # Use maximum within each group for first grouping
            df = _group_exceedances(self, df, groupby_cols[0]+[self.stat_col])
        
        # Count exceedances
        if groupby_cols == None:
            df = _percent_exceedances(self, df,
                                      threshold, threshold_is_exceedance,
                                      [self.stat_col],
                                      round_to)
        else:
            df = _percent_exceedances(self, df,
                                      threshold, threshold_is_exceedance,
                                      groupby_cols[-1]+[self.stat_col],
                                      round_to)
        
        # Reorder columns
        df[self.result_col] = df.pop(self.result_col)
        df[self.numeric_col] = df.pop(self.numeric_col)
        df[self.exceedances_col] = df.pop(self.exceedances_col)
        df[self.non_exceedances_col] = df.pop(self.non_exceedances_col)
        df[self.ignored_col] = df.pop(self.ignored_col)
        
        # Merge count info
        if count_cols:
            df = _merge_count_info(self, df, groupby_cols, count_cols, filters)
        
        return df



'''
The function in this module compares intervals to a threshold and determines
the percentage of values that exceed the threshold. Required information
includes (i) a CensoredData input object that contains the original data table
as well as column names, (ii) information on the threshold, (iii) the
specific column groupings used to combine data, and (iv) additional settings

'''

import numpy as np

def _determine_exceedances(cdf,
                           df,
                           threshold,
                           threshold_is_exceedance):
    
    # True or False are considered integer types so check for boolean too
    if (isinstance(threshold, bool) |
            (not isinstance(threshold,(int, float)))):
        raise ValueError(f'The value supplied to threshold must be numeric.'
            'Instead, the value supplied was '
            f'{repr(threshold)}.')
    
    # Check that threshold_is_exceedance is boolean
    if not isinstance(threshold_is_exceedance, bool):
        raise ValueError(f'The value provided for threshold_is_exceedance '
            'must be True or False. The value provided was '
            f'{threshold_is_exceedance}')
    
    # Determine exceedance (1=exceedance, 0=non-exceedance, nan=indeterminate)
    
    # Set conditions depending on whether the threshold is an exceedance
    if threshold_is_exceedance:
        # Set conditions for exceedances and non-exceedances
        conditions = [
            # Exceedance condition
            # The left bound is greater than or equal to the threshold
            (df[cdf.left_bound_col] >= threshold),
            # Non-exceedance conditions
            (
                # The right bound is less than the threshold
                (df[cdf.right_bound_col] < threshold) |
                # The right bound is equal to the threshold and open
                ((df[cdf.right_bound_col] == threshold) &
                 (df[cdf.right_boundary_col] == 'Open'))
            )
            ]
    # Set conditions when threshold is not an exceedance
    else:
        # Set conditions for exceedances and non-exceedances
        conditions = [
            # Exceedances conditions
            (
                # The left bound is greater than the threshold
                (df[cdf.left_bound_col] > threshold) |
                # If the left bound is equal to the threshold and open
                ((df[cdf.left_bound_col] == threshold) &
                 (df[cdf.left_boundary_col] == 'Open'))
            ),
            # Non-exceedance condition
            # The right bound is less than or equal to the threshold
            (df[cdf.right_bound_col] <= threshold)
            ]
    # Exceedances are indicated by integers
    # (1=exceedance, 0=non-exceedance)
    results = [
        1,
        0
        ]
    # Create column with exceedance results
    df[cdf.exceedances_col] = np.select(conditions, results, np.nan)
    
    return df
    
def _group_exceedances(cdf,
                       df,
                       groupby_cols):
    
    # Take maximum within each group
    df = (df.groupby(groupby_cols)[cdf.exceedances_col]
              .max()
              .reset_index())
    
    return df

def _percent_exceedances(cdf,
                         df,
                         threshold,
                         threshold_is_exceedance,
                         groupby_cols,
                         round_to):
    
    # Within groups, count exceedances, determinate values, and all values
    # including indeterminate
    df = (df.groupby(groupby_cols)
          .agg(**{
              cdf.exceedances_col: (cdf.exceedances_col,'sum'),
              '__DeterminedCount__': (cdf.exceedances_col,'count'),
              '__TotalCount__': (cdf.exceedances_col,'size')
            })
        )
    
    # Determine counts of non-exceedances and ignored/dropped values
    df[cdf.non_exceedances_col] = (
        df['__DeterminedCount__'] - df[cdf.exceedances_col]
        )
    df[cdf.ignored_col] = df['__TotalCount__'] - df['__DeterminedCount__']
    
    # Calculate percentage and round
    df[cdf.numeric_col] = (
        round(df[cdf.exceedances_col] / df['__DeterminedCount__'] * 100,
              round_to)
        )
    # Create result and included count of dropped values
    df[cdf.result_col] = df[cdf.numeric_col].astype(str) + '%'
    df.loc[df[cdf.ignored_col] > 0, cdf.result_col] = (
        df[cdf.result_col] +
        ' (' + df[cdf.ignored_col].astype(str) + ' value(s) ignored)'
        )
    
    # Include threshold in output
    df[cdf.threshold_col] = f'{threshold}'
    if threshold_is_exceedance:
        df[cdf.threshold_col] = '<' + df[cdf.threshold_col]
    
    # Reset index
    df = df.reset_index()
    
    # Drop working columns
    df = df.drop(df.filter(regex='__').columns, axis=1)
    
    return df

    
